// All CUDA specific stuff goes here
//

#include <cuda.h>
#include <cutil.h>
#include "defs.h"


#define CHECK_BANK_CONFLICTS 0
#if CHECK_BANK_CONFLICTS	
#define AS(i, j) cutilBankChecker(((float*)&As[0][0]), (32 * i + j))
#define BS(i, j) cutilBankChecker(((float*)&Bs[0][0]), (10 * i + j))
	#else
#define AS(i, j) As[i][j]
#define BS(i, j) Bs[i][j]
#endif

extern "C" float applySignalGPU(Matrix, Signals, Signals);
extern "C" float applySignalCPU(Matrix, Signals, Signals);
extern "C" int checkResults(Signals, Signals);

void kernelApplySignal(Matrix, Signals, Signals);

// performs matrix vector multiplication
// Note: _input, _result are pointer references to where data is read/written
void ApplySignal(Matrix Mat, Signals _input, Signals _result)
{
    int i, j;
    for(i = 0; i < MAT_ROW_DIM; i++)
    {
        _result[i] = 0.0;
        for(j = 0; j < MAT_COL_DIM; j++)
        {
            _result[i] += Mat[i * MAT_COL_DIM + j] * _input[j];
        }
    } 
printf("In CPU ApplySignal");   
}

/*  C Interface function called from matvec.c
 *  Mat : Input Transformation Matrix
 *  _ipsignals : Array of input Signals
 *  _cpuresults : Array of Signal results computed on CPU
 *  returns the elapsed time to compute on CPU
 */
float applySignalCPU(Matrix Mat, Signals _ipsignals, Signals _cpuresults)
{
    int i, j;
    // create CUDA event handles for timing purposes
    cudaEvent_t start_event, stop_event;
    float seq_elapsed_time;

    CUDA_SAFE_CALL(cudaEventCreate(&start_event));
    CUDA_SAFE_CALL(cudaEventCreate(&stop_event));

    // start the timer for seq code
    cudaEventRecord(start_event, 0);

    // pointer referennces to large signal arrays 'signals', 'resultSignals'
    Signals _input, _result;

    _input = _ipsignals;
    _result = _cpuresults;

    for(i = 0; i < NSEQ; i++) {
        _input = (i * VEC_DIM) + _ipsignals;
        _result = (i * MAT_ROW_DIM) + _cpuresults;
        ApplySignal(Mat, _input, _result);
    }

    // stop timing
    cudaEventRecord(stop_event, 0);
    cudaEventSynchronize(stop_event);    

    CUDA_SAFE_CALL(cudaEventElapsedTime(&seq_elapsed_time, start_event, stop_event));

    return seq_elapsed_time; 
}

Signals allocatematrix()
{
    Signals transB;
    transB = (Signals) malloc(sizeof(float) * NSEQ * VEC_DIM);

    if(transB == NULL) {
        ERROR_MSG("error: memory allocation failed for signal array\n")
    }
    return transB;
}

//global function
__global__ void kernelApplySignal(Signals A,Matrix B,Signals C){

    // Block index
    int bx = blockIdx.x;
    int by = blockIdx.y;

    // Thread index
    int tx = threadIdx.x;
    int ty = threadIdx.y;

    // Index of the first sub-matrix of A processed by the block
    int aBegin = MAT_COL_DIM * 10 * by;	

    // Index of the last sub-matrix of A processed by the block
    int aEnd   = aBegin + MAT_COL_DIM - 1;		

    // Step size used to iterate through the sub-matrices of A
    int aStep  = 32;			

    // Index of the first sub-matrix of A processed by the block
    
    int bBegin = 10 * bx;	

    // Step size used to iterate through the sub-matrices of B
    int bStep  = 32 * NSEQ;

    // Csub is used to store the element of the block sub-matrix
    // that is computed by the thread

    float Csub = 0;

    // Loop over all the sub-matrices of A and B
    // required to compute the block sub-matrix
    for (int a = aBegin, b = bBegin;
             a <= aEnd;
             a += aStep, b += bStep) {

        // Declaration of the shared memory array As used to
        // store the sub-matrix of A
        __shared__ float As[10][32];

        // Declaration of the shared memory array Bs used to
        // store the sub-matrix of B
        __shared__ float Bs[32][10];

        // Load the matrices from device memory
        // to shared memory; each thread loads
        // one element of each matrix
        AS(ty, tx) = A[a + MAT_COL_DIM * ty + tx];
        BS(ty, tx) = B[b + NSEQ * ty + tx];

        // Synchronize to make sure the matrices are loaded
        __syncthreads();

        // Multiply the two matrices together;
        // each thread computes one element
        // of the block sub-matrix
        for (int k = 0; k < 32; ++k){
		
	     // Using global Memory
              Csub += A[a + MAT_COL_DIM * ty +k] * B[b + VEC_DIM * ty + k];
	     // Using shared Memory
	        //Csub += AS(ty,k) * BS(k,tx);
	     // Using Registers
		//float reg_A = AS(ty,k);
		//float reg_B = BS(ty,k);
		//Csub+=reg_A*reg_B;
	    }
         // Synchronize to make sure that the preceding
         // computation is done before loading two new
         // sub-matrices of A and B in the next iteration
         __syncthreads();
    }
    // Write the block sub-matrix to device memory;
    // each thread writes one element
    int c = NSEQ * 32 * by + 10 * bx;
    C[c + NSEQ * ty + tx] = Csub;
    }

/*  C Interface function called from matvec.c
 *  Mat : Input Transformation Matrix
 *  _ipsignals : Array of input Signals
 *  _gpuresults : Array of Signal results computed on GPU
 *  returns the elapsed time to compute on GPU
 */

float applySignalGPU(Matrix Mat, Signals _ipsignal, Signals _gpuresults)
{
    int i,j;
    cudaEvent_t start_event, stop_event;
    float cuda_elapsed_time;

    // CUDA Memory allocation and initialization
    CUDA_SAFE_CALL( cudaEventCreate(&start_event));
    CUDA_SAFE_CALL( cudaEventCreate(&stop_event));

    // start the timer for GPU code
    cudaEventRecord(start_event, 0);

    /*Allocate memory in device for Matrix and Signal Vector*/
    unsigned int size_M = MAT_ROW_DIM * MAT_COL_DIM;
    unsigned int mem_size_M = sizeof(float) * size_M;
    Matrix d_M;
    cudaMalloc((void**) &d_M, mem_size_M);

    unsigned int size_S = VEC_DIM * NSEQ;
    unsigned int mem_size_S = sizeof(float) * size_S;
    Matrix d_S;
    cudaMalloc((void**) &d_S, mem_size_S);

    
    /*Allocate memory for the resultant vector*/
    unsigned int size_R = VEC_DIM * NSEQ;
    unsigned int mem_size_R = sizeof(float) * size_R;
    //float* d_R;
    Matrix d_R;
    cudaMalloc((void**) &d_R, mem_size_R);


    /*Copy matrix and vector from host to device*/
    cudaMemcpy(d_M, Mat, mem_size_M,cudaMemcpyHostToDevice);
    cudaMemcpy(d_S, _ipsignal, mem_size_S,cudaMemcpyHostToDevice);
    
    /*setup execution parameters*/
    dim3 threads(10,32);
    dim3 grid(1,160);
    
    kernelApplySignal<<<grid,threads>>>(d_M,d_S,d_R);
    printf("Returned from kernel \n");
    cudaThreadSynchronize();
    // stop timing
    cudaEventRecord(stop_event, 0);
    cudaEventSynchronize(stop_event);    
    printf("CUDA Elapsed Time:\t%.2f ms\n", cuda_elapsed_time);
    CUDA_SAFE_CALL(cudaEventElapsedTime(&cuda_elapsed_time, start_event, stop_event));
    printf("Returned from Device \n");
     
    /*Copying back from Device to Host*/
    cudaMemcpy(_gpuresults,d_R,mem_size_R,cudaMemcpyDeviceToHost);
    printf("Copied back from device to Host \n ");
    //free(d_M);
    //free(d_S);
    //free(d_R);
    return cuda_elapsed_time;
}

// Checks for memory equivalence
int checkResults(Signals _cpuresult, Signals _gpuresult)
{
    // compares two float pointer references for equivalence
	CUTBoolean res = cutComparefe((float*)_cpuresult, (float*)_gpuresult, NSEQ * MAT_ROW_DIM, 0.0005f);
    // if the test passes, print CPU/GPU time and cleanup

    return (1 == res);
}

